@@ -1,22 +1,23 @@
 /*
- *  Copyright 2001 Adrian Thurston <thurston@complang.org>
- */
-
-/*  This file is part of Aapl.
+ * Copyright 2001 Adrian Thurston <thurston@colm.net>
  *
- *  Aapl is free software; you can redistribute it and/or modify it under the
- *  terms of the GNU Lesser General Public License as published by the Free
- *  Software Foundation; either version 2.1 of the License, or (at your option)
- *  any later version.
+ * Permission is hereby granted, free of charge, to any person obtaining a copy
+ * of this software and associated documentation files (the "Software"), to
+ * deal in the Software without restriction, including without limitation the
+ * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
+ * sell copies of the Software, and to permit persons to whom the Software is
+ * furnished to do so, subject to the following conditions:
  *
- *  Aapl is distributed in the hope that it will be useful, but WITHOUT ANY
- *  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
- *  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
- *  more details.
+ * The above copyright notice and this permission notice shall be included in all
+ * copies or substantial portions of the Software.
  *
- *  You should have received a copy of the GNU Lesser General Public License
- *  along with Aapl; if not, write to the Free Software Foundation, Inc., 59
- *  Temple Place, Suite 330, Boston, MA 02111-1307 USA
+ * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
+ * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
+ * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
+ * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
+ * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
+ * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
+ * SOFTWARE.
  */
 
 /* This header is not wrapped in ifndef becuase it is not intended to
@@ -883,7 +884,7 @@ template <AVLMEL_TEMPDEF> Element *AvlTree<AVLMEL_TEMPUSE>::
 #ifdef AVL_BASIC
 		keyRelation = this->compare( *element, *curEl );
 #else
-		keyRelation = this->compare( element->BASEKEY(getKey()), 
+		keyRelation = this->compare( element->BASEKEY(getKey()),
 				curEl->BASEKEY(getKey()) );
 #endif
 
@@ -1295,6 +1296,8 @@ template <AVLMEL_TEMPDEF> void AvlTree<AVLMEL_TEMPUSE>::empty()
 
 #ifdef WALKABLE
 		BASELIST::abandon();
+#else
+		head = tail = 0;
 #endif
 	}
 }
@@ -1311,6 +1314,8 @@ template <AVLMEL_TEMPDEF> void AvlTree<AVLMEL_TEMPUSE>::abandon()
 
 #ifdef WALKABLE
 	BASELIST::abandon();
+#else
+	head = tail = 0;
 #endif
 }
 
