@@ -1,5 +1,5 @@
 /*
- * Copyright 2001-2014 Adrian Thurston <thurston@colm.net>
+ * Copyright 2001-2018 Adrian Thurston <thurston@colm.net>
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to
@@ -54,11 +54,35 @@ struct LongestMatchPart;
 
 string itoa( int i );
 
+struct Variable
+{
+	Variable( const char *name ) : name(name), isReferenced(false) {}
+
+	operator const std::string() { isReferenced = true; return name; }
+	void reference() { isReferenced = true; }
+
+	const char *name;
+	bool isReferenced;
+};
+
+struct GotoLabel
+{
+	GotoLabel( const char *name ) : name(name), isReferenced(false) {}
+
+	operator std::string() { isReferenced = true; return name; }
+	void reference() { isReferenced = true; }
+
+	const char *name;
+	bool isReferenced;
+};
+
+std::ostream &operator<<( std::ostream &out, GotoLabel &l );
+std::ostream &operator<<( std::ostream &out, Variable &v );
+
 struct TableArray;
 typedef Vector<TableArray*> ArrayVector;
 struct CodeGen;
 
-
 struct TableArray
 {
 	enum State {
@@ -80,7 +104,7 @@ struct TableArray
 		this->isChar = isChar;
 	}
 
-	std::string ref() const;
+	std::string ref();
 
 	void value( long long v );
 
@@ -111,6 +135,8 @@ struct TableArray
 	CodeGen &codeGen;
 	std::ostream &out;
 	int ln;
+	bool isReferenced;
+	bool started;
 };
 
 struct IlOpts
@@ -145,21 +171,33 @@ protected:
 	typedef Vector<TableArray*> ArrayVector;
 	ArrayVector arrayVector;
 
+	Variable cpc;
+	Variable pop_test;
+	Variable new_recs;
+	Variable alt;
+
 	string FSM_NAME();
 	string START_STATE_ID();
 	void taActions();
-	string TABS( int level );
 	string KEY( Key key );
 	string LDIR_PATH( char *path );
 
 	void ACTION( ostream &ret, GenAction *action, IlOpts opts );
 	void NFA_CONDITION( ostream &ret, GenAction *condition, bool last );
+	void NFA_POP_TEST_EXEC();
 	void CONDITION( ostream &ret, GenAction *condition );
 	string ALPH_TYPE();
 
 	bool isAlphTypeSigned();
 	long long tableData;
+	RagelBackend backend;
 	bool stringTables;
+	BackendFeature backendFeature;
+
+	TableArray nfaTargs;
+	TableArray nfaOffsets;
+	TableArray nfaPushActions;
+	TableArray nfaPopTrans;
 
 	virtual string GET_KEY();
 
@@ -180,20 +218,21 @@ protected:
 	string ERROR() { return DATA_PREFIX() + "error"; }
 	string FIRST_FINAL() { return DATA_PREFIX() + "first_final"; }
 
+	/* Declare a variable only if referenced. */
+	void DECLARE( std::string type, Variable &var, std::string init = "" );
+
 	string CAST( string type );
 
 	string ARR_TYPE( const TableArray &ta )
 		{ return ta.type; }
 
-	string ARR_REF( const TableArray &ta )
+	string ARR_REF( TableArray &ta )
 		{ return ta.ref(); }
 
 	void INLINE_EXPR( ostream &ret, GenInlineList *inlineList );
 	void INLINE_BLOCK( ostream &ret, GenInlineExpr *inlineExpr );
 	void INLINE_PLAIN( ostream &ret, GenInlineExpr *inlineExpr );
 
-	void EOF_CHECK( ostream &ret );
-
 	void INLINE_LIST( ostream &ret, GenInlineList *inlineList, 
 			int targState, bool inFinish, bool csForced );
 	virtual void GOTO( ostream &ret, int gotoDest, bool inFinish ) = 0;
@@ -236,32 +275,33 @@ protected:
 
 	string STR( int v );
 
-	bool outLabelUsed;
-	bool testEofUsed;
-	bool againLabelUsed;
-	bool useIndicies;
-	bool matchCondLabelUsed;
-
 	void VALUE( string type, string name, string value );
 
 	string ACCESS_OPER()
-		{ return ""; }
+		{ return backend == Direct ? "" : " -> "; }
 
 	string OPEN_HOST_EXPR()
-		{ return "("; }
+		{ return backend == Direct ? "(" : "host( \"-\", 1 ) ={"; }
 
 	string OPEN_HOST_EXPR( string fileName, int line )
-		{ return "("; }
+	{
+		return backend == Direct ? "(" : "host( \"" + fileName + "\", " + STR(line) + " ) ={"; 
+	}
 
 	string CLOSE_HOST_EXPR()
-		{ return ")"; }
+		{ return backend == Direct ? ")" : "}="; }
 
 	string OPEN_HOST_BLOCK( string fileName, int line )
 	{ 
-		if ( lineDirectives )
-			return "{\n#line " + STR(line) + " \"" + fileName + "\"\n";
-		else
-			return "{\n";
+		if ( backend == Direct ) {
+			std::stringstream ss;
+			ss << "{\n" ;
+			(*genLineDirective)( ss, lineDirectives, line, fileName.c_str() );
+			return ss.str();
+		}
+		else {
+			return "host( \"" + fileName + "\", " + STR(line) + " ) ${";
+		}
 	}
 
 	string OPEN_HOST_BLOCK( GenInlineExpr *inlineExpr )
@@ -270,43 +310,52 @@ protected:
 	}
 
 	string CLOSE_HOST_BLOCK()
-		{ return "}"; }
+		{ return backend == Direct ? "}\n" : "}$"; }
 
 	string OPEN_HOST_PLAIN()
-		{ return ""; }
+		{ return backend == Direct ? "" : "host( \"-\", 1 ) @{"; }
 
 	string CLOSE_HOST_PLAIN()
-		{ return ""; }
+		{ return backend == Direct ? "" : "}@"; }
 
 	string OPEN_GEN_EXPR()
-		{ return "("; }
+		{ return backend == Direct ? "(" : "={"; }
 
 	string CLOSE_GEN_EXPR()
-		{ return ")"; }
+		{ return backend == Direct ? ")" : "}="; }
 
 	string OPEN_GEN_BLOCK()
-		{ return "{"; }
+		{ return backend == Direct ? "{" : "${"; }
 
 	string CLOSE_GEN_BLOCK()
-		{ return "}"; }
+		{ return backend == Direct ? "}" : "}$"; }
 
 	string OPEN_GEN_PLAIN()
-		{ return ""; }
+		{ return backend == Direct ? "" : "@{"; }
 
 	string CLOSE_GEN_PLAIN()
-		{ return ""; }
+		{ return backend == Direct ? "" : "}@"; }
 	
+	string INT()
+		{ return "int"; }
+
 	string UINT()
-		{ return "unsigned int"; }
+		{ return backend == Direct ? "unsigned int" : "uint"; }
 
 	string INDEX( string type, string name )
 	{
-		return "const " + type + " *" + name;
+		if ( backend == Direct )
+			return "const " + type + " *" + name;
+		else
+			return "index " + type + " " + name;
 	}
 
-	string ENTRY()
+	string INDEX( string type )
 	{
-		return "";
+		if ( backend == Direct )
+			return "const " + type + " *";
+		else
+			return "index " + type + " ";
 	}
 
 	string LABEL( string name )
@@ -314,51 +363,97 @@ protected:
 		return name + ": ";
 	}
 
+	string EMIT_LABEL( GotoLabel label )
+	{
+		if ( label.isReferenced )
+			return std::string(label.name) + ": {}\n";
+		else
+			return "";
+	}
+
 	string OFFSET( string arr, string off )
 	{
-		return "( " + arr + " + (" + off + "))";
+		if ( backend == Direct )
+			return "( " + arr + " + (" + off + "))";
+		else
+			return "offset( " + arr + ", " + off + " )";
 	}
 
 	string TRUE()
 	{
-		return "1";
+		if ( backend == Direct )
+			return "1";
+		else
+			return "TRUE";
 	}
 
 	string DEREF( string arr, string off )
 	{
-		return "(*( " + off + "))";
+		if ( backend == Direct )
+			return "(*( " + off + "))";
+		else
+			return "deref( " + arr + ", " + off + " )";
 	}
 	
 	string CASE( string val )
 	{
-		return "case " + val + ": ";
+		if ( backend == Direct )
+			return "case " + val + ": ";
+		else
+			return "case " + val;
 	}
 
 	string DEFAULT()
 	{
-		return "default:";
+		if ( backend == Direct )
+			return "default:";
+		else
+			return "default";
 	}
 
 	string CEND( )
 	{
-		return " break; ";
+		if ( backend == Direct )
+			return " break; ";
+		else
+			return " ";
 	}
 
 	string FALLTHROUGH()
 	{
-		return " ";
+		if ( backend == Direct )
+			return " ";
+		else
+			return "fallthrough;";
 	}
 
 	string NIL()
 	{
-		return "0";
+		if ( backend == Direct )
+			return "0";
+		else
+			return "nil";
 	}
 
 	string EXPORT( string type, string name, string value )
 	{
-		return "#define " + name + " " + value;
+		if ( backend == Direct )
+			return "#define " + name + " " + value;
+		else
+			return "export " + type + " " + name + " " + value + ";";
 	}
 
+	void NFA_POST_POP();
+	virtual void NFA_PUSH( std::string );
+	virtual void NFA_POP() = 0;
+	virtual void LOCATE_TRANS() {}
+	virtual void LOCATE_COND() {}
+	virtual void EOF_TRANS() {}
+
+
+	virtual void COND_EXEC( std::string expr ) {}
+	virtual void COND_BIN_SEARCH( Variable &var, TableArray &keys, std::string ok, std::string error ) {}
+
 public:
 	virtual void writeExports();
 };
