@@ -1,5 +1,5 @@
 /*
- * Copyright 2001-2014 Adrian Thurston <thurston@colm.net>
+ * Copyright 2001-2018 Adrian Thurston <thurston@colm.net>
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to
@@ -30,18 +30,18 @@
 
 using std::ostringstream;
 
-Goto::Goto( const CodeGenArgs &args ) 
-:
-	CodeGen( args ),
-	actions(           "actions",             *this ),
-	toStateActions(    "to_state_actions",    *this ),
-	fromStateActions(  "from_state_actions",  *this ),
-	eofActions(        "eof_actions",         *this ),
-	nfaTargs(          "nfa_targs",           *this ),
-	nfaOffsets(        "nfa_offsets",         *this ),
-	nfaPushActions(    "nfa_push_actions",    *this ),
-	nfaPopTrans(       "nfa_pop_trans",        *this )
-{}
+IpLabel *Goto::allocateLabels( IpLabel *labels, IpLabel::Type type, int n )
+{
+	if ( labels == 0 ) {
+		labels = new IpLabel[n];
+		for ( int id = 0; id < n; id++ ) {
+			labels[id].type = type;
+			labels[id].stid = id;
+		}
+	}
+
+	return labels;
+}
 
 void Goto::setTableState( TableArray::State state )
 {
@@ -51,16 +51,15 @@ void Goto::setTableState( TableArray::State state )
 	}
 }
 
-
 /* Emit the goto to take for a given transition. */
-std::ostream &Goto::COND_GOTO( RedCondPair *cond, int level )
+std::ostream &Goto::COND_GOTO( RedCondPair *cond )
 {
-	out << TABS(level) << "goto ctr" << cond->id << ";";
+	out << "goto " << ctrLabel[cond->id].reference() << ";";
 	return out;
 }
 
 /* Emit the goto to take for a given transition. */
-std::ostream &Goto::TRANS_GOTO( RedTransAp *trans, int level )
+std::ostream &Goto::TRANS_GOTO( RedTransAp *trans )
 {
 	if ( trans->condSpace == 0 || trans->condSpace->condSet.length() == 0 ) {
 		/* Existing. */
@@ -68,22 +67,22 @@ std::ostream &Goto::TRANS_GOTO( RedTransAp *trans, int level )
 		RedCondPair *cond = trans->outCond( 0 );
 
 		/* Go to the transition which will go to the state. */
-		out << TABS(level) << "goto ctr" << cond->id << ";";
+		out << "goto " << ctrLabel[cond->id].reference() << ";";
 	}
 	else {
-		out << TABS(level) << "int ck = 0;\n";
+		out << ck << " = 0;\n";
 		for ( GenCondSet::Iter csi = trans->condSpace->condSet; csi.lte(); csi++ ) {
-			out << TABS(level) << "if ( ";
+			out << "if ( ";
 			CONDITION( out, *csi );
 			Size condValOffset = (1 << csi.pos());
-			out << " ) ck += " << condValOffset << ";\n";
+			out << " )\n" << ck << " += " << condValOffset << ";\n";
 		}
 		CondKey lower = 0;
 		CondKey upper = trans->condFullSize() - 1;
-		COND_B_SEARCH( trans, 1, lower, upper, 0, trans->numConds()-1 );
+		COND_B_SEARCH( trans, lower, upper, 0, trans->numConds()-1 );
 
 		if ( trans->errCond() != 0 ) {
-			COND_GOTO( trans->errCond(), level+1 ) << "\n";
+			COND_GOTO( trans->errCond() ) << "\n";
 		}
 	}
 
@@ -108,158 +107,6 @@ void Goto::taActions()
 	actions.finish();
 }
 
-void Goto::NFA_PUSH()
-{
-	if ( redFsm->anyNfaStates() ) {
-		out <<
-			"	if ( " << ARR_REF( nfaOffsets ) << "[" << vCS() << "] ) {\n"
-			"		int alt = 0; \n"
-			"		int new_recs = " << ARR_REF( nfaTargs ) << "[" << CAST("int") <<
-						ARR_REF( nfaOffsets ) << "[" << vCS() << "]];\n";
-
-		if ( red->nfaPrePushExpr != 0 ) {
-			out << OPEN_HOST_BLOCK( red->nfaPrePushExpr );
-			INLINE_LIST( out, red->nfaPrePushExpr->inlineList, 0, false, false );
-			out << CLOSE_HOST_BLOCK();
-		}
-
-		out <<
-			"		while ( alt < new_recs ) { \n";
-
-
-		out <<
-			"			nfa_bp[nfa_len].state = " << ARR_REF( nfaTargs ) << "[" << CAST("int") <<
-							ARR_REF( nfaOffsets ) << "[" << vCS() << "] + 1 + alt];\n"
-			"			nfa_bp[nfa_len].p = " << P() << ";\n";
-
-		if ( redFsm->bAnyNfaPops ) {
-			out <<
-				"			nfa_bp[nfa_len].popTrans = " << CAST("long") <<
-								ARR_REF( nfaOffsets ) << "[" << vCS() << "] + 1 + alt;\n"
-				"\n"
-				;
-		}
-
-		if ( redFsm->bAnyNfaPushes ) {
-			out <<
-				"			switch ( " << ARR_REF( nfaPushActions ) << "[" << CAST("int") <<
-								ARR_REF( nfaOffsets ) << "[" << vCS() << "] + 1 + alt] ) {\n";
-
-			/* Loop the actions. */
-			for ( GenActionTableMap::Iter redAct = redFsm->actionMap;
-					redAct.lte(); redAct++ )
-			{
-				if ( redAct->numNfaPushRefs > 0 ) {
-					/* Write the entry label. */
-					out << "\t " << CASE( STR( redAct->actListId+1 ) ) << " {\n";
-
-					/* Write each action in the list of action items. */
-					for ( GenActionTable::Iter item = redAct->key; item.lte(); item++ )
-						ACTION( out, item->value, IlOpts( 0, false, false ) );
-
-					out << "\n\t" << CEND() << "}\n";
-				}
-			}
-
-			out <<
-				"			}\n";
-		}
-
-
-		out <<
-			"			nfa_len += 1;\n"
-			"			alt += 1;\n"
-			"		}\n"
-			"	}\n"
-			;
-	}
-}
-
-void Goto::NFA_POP()
-{
-	if ( redFsm->anyNfaStates() ) {
-		out <<
-			"	if ( nfa_len > 0 ) {\n";
-
-		if ( redFsm->bAnyNfaCondRefs )
-			out << "	int _cpc;\n";
-
-		out <<
-			"		nfa_count += 1;\n"
-			"		nfa_len -= 1;\n"
-			"		" << P() << " = nfa_bp[nfa_len].p;\n"
-			;
-
-		if ( redFsm->bAnyNfaPops ) {
-			out << 
-				"		int _pop_test = 1;\n"
-				"		switch ( " << ARR_REF( nfaPopTrans ) <<
-							"[nfa_bp[nfa_len].popTrans] ) {\n";
-
-			/* Loop the actions. */
-			for ( GenActionTableMap::Iter redAct = redFsm->actionMap;
-					redAct.lte(); redAct++ )
-			{
-				if ( redAct->numNfaPopTestRefs > 0 ) {
-					/* Write the entry label. */
-					out << "\t " << CASE( STR( redAct->actListId+1 ) ) << " {\n";
-
-					/* Write each action in the list of action items. */
-					for ( GenActionTable::Iter item = redAct->key; item.lte(); item++ )
-						NFA_CONDITION( out, item->value, item.last() );
-
-					out << "\n\t" << CEND() << "}\n";
-				}
-			}
-
-			out <<
-				"		}\n";
-
-			out <<
-				"		if ( _pop_test ) {\n"
-				"			" << vCS() << " = nfa_bp[nfa_len].state;\n";
-
-			if ( red->nfaPostPopExpr != 0 ) {
-				out << OPEN_HOST_BLOCK( red->nfaPostPopExpr );
-				INLINE_LIST( out, red->nfaPostPopExpr->inlineList, 0, false, false );
-				out << CLOSE_HOST_BLOCK();
-			}
-
-			out <<
-				"			goto _resume;\n"
-				"		}\n";
-
-			if ( red->nfaPostPopExpr != 0 ) {
-				out <<
-				"			else {\n"
-				"			" << OPEN_HOST_BLOCK( red->nfaPostPopExpr );
-				INLINE_LIST( out, red->nfaPostPopExpr->inlineList, 0, false, false );
-				out << CLOSE_HOST_BLOCK() << "\n"
-				"			}\n";
-			}
-		}
-		else {
-			out <<
-				"		" << vCS() << " = nfa_bp[nfa_len].state;\n";
-
-			if ( red->nfaPostPopExpr != 0 ) {
-				out << OPEN_HOST_BLOCK( red->nfaPostPopExpr );
-				INLINE_LIST( out, red->nfaPostPopExpr->inlineList, 0, false, false );
-				out << CLOSE_HOST_BLOCK();
-			}
-
-			out <<
-				"		goto _resume;\n";
-		}
-
-		out << 
-			"		goto _out;\n"
-			"	}\n";
-	}
-}
-
-
-
 void Goto::GOTO_HEADER( RedStateAp *state )
 {
 	/* Label the state. */
@@ -275,30 +122,30 @@ void Goto::SINGLE_SWITCH( RedStateAp *state )
 
 	if ( numSingles == 1 ) {
 		/* If there is a single single key then write it out as an if. */
-		out << "\tif ( " << GET_KEY() << " == " << 
-				KEY(data[0].lowKey) << " ) {\n\t\t"; 
+		out << "if ( " << GET_KEY() << " == " << 
+				KEY(data[0].lowKey) << " ) {\n"; 
 
 		/* Virtual function for writing the target of the transition. */
-		TRANS_GOTO(data[0].value, 0) << "\n";
-		out << "\t}\n";
+		TRANS_GOTO(data[0].value) << "\n";
+		out << "}\n";
 	}
 	else if ( numSingles > 1 ) {
 		/* Write out single keys in a switch if there is more than one. */
-		out << "\tswitch( " << GET_KEY() << " ) {\n";
+		out << "switch( " << GET_KEY() << " ) {\n";
 
 		/* Write out the single indicies. */
 		for ( int j = 0; j < numSingles; j++ ) {
-			out << "\t\t case " << KEY(data[j].lowKey) << ": {\n";
-			TRANS_GOTO(data[j].value, 0) << "\n";
-			out << "\t}\n";
+			out << "case " << KEY(data[j].lowKey) << ": {\n";
+			TRANS_GOTO(data[j].value) << "\n";
+			out << "}\n";
 		}
 		
 		/* Close off the transition switch. */
-		out << "\t}\n";
+		out << "}\n";
 	}
 }
 
-void Goto::RANGE_B_SEARCH( RedStateAp *state, int level, Key lower, Key upper, int low, int high )
+void Goto::RANGE_B_SEARCH( RedStateAp *state, Key lower, Key upper, int low, int high )
 {
 	/* Get the mid position, staying on the lower end of the range. */
 	int mid = (low + high) >> 1;
@@ -314,83 +161,83 @@ void Goto::RANGE_B_SEARCH( RedStateAp *state, int level, Key lower, Key upper, i
 
 	if ( anyLower && anyHigher ) {
 		/* Can go lower and higher than mid. */
-		out << TABS(level) << "if ( " << GET_KEY() << " < " << 
+		out << "if ( " << GET_KEY() << " < " << 
 				KEY(data[mid].lowKey) << " ) {\n";
-		RANGE_B_SEARCH( state, level+1, lower, keyOps->sub( data[mid].lowKey, 1 ), low, mid-1 );
-		out << TABS(level) << "} else if ( " << GET_KEY() << " > " << 
+		RANGE_B_SEARCH( state, lower, keyOps->sub( data[mid].lowKey, 1 ), low, mid-1 );
+		out << "} else if ( " << GET_KEY() << " > " << 
 				KEY(data[mid].highKey) << " ) {\n";
-		RANGE_B_SEARCH( state, level+1, keyOps->add( data[mid].highKey, 1 ), upper, mid+1, high );
-		out << TABS(level) << "} else {\n";
-		TRANS_GOTO(data[mid].value, level+1) << "\n";
-		out << TABS(level) << "}\n";
+		RANGE_B_SEARCH( state, keyOps->add( data[mid].highKey, 1 ), upper, mid+1, high );
+		out << "} else {\n";
+		TRANS_GOTO(data[mid].value) << "\n";
+		out << "}\n";
 	}
 	else if ( anyLower && !anyHigher ) {
 		/* Can go lower than mid but not higher. */
-		out << TABS(level) << "if ( " << GET_KEY() << " < " << 
+		out << "if ( " << GET_KEY() << " < " << 
 				KEY(data[mid].lowKey) << " ) {\n";
-		RANGE_B_SEARCH( state, level+1, lower, keyOps->sub( data[mid].lowKey, 1 ), low, mid-1 );
+		RANGE_B_SEARCH( state, lower, keyOps->sub( data[mid].lowKey, 1 ), low, mid-1 );
 
 		/* if the higher is the highest in the alphabet then there is no
 		 * sense testing it. */
 		if ( limitHigh ) {
-			out << TABS(level) << "} else {\n";
-			TRANS_GOTO(data[mid].value, level+1) << "\n";
-			out << TABS(level) << "}\n";
+			out << "} else {\n";
+			TRANS_GOTO(data[mid].value) << "\n";
+			out << "}\n";
 		}
 		else {
-			out << TABS(level) << "} else if ( " << GET_KEY() << " <= " << 
+			out << "} else if ( " << GET_KEY() << " <= " << 
 					KEY(data[mid].highKey) << " ) {\n";
-			TRANS_GOTO(data[mid].value, level+1) << "\n";
-			out << TABS(level) << "}\n";
+			TRANS_GOTO(data[mid].value) << "\n";
+			out << "}\n";
 		}
 	}
 	else if ( !anyLower && anyHigher ) {
 		/* Can go higher than mid but not lower. */
-		out << TABS(level) << "if ( " << GET_KEY() << " > " << 
+		out << "if ( " << GET_KEY() << " > " << 
 				KEY(data[mid].highKey) << " ) {\n";
-		RANGE_B_SEARCH( state, level+1, keyOps->add( data[mid].highKey, 1 ), upper, mid+1, high );
+		RANGE_B_SEARCH( state, keyOps->add( data[mid].highKey, 1 ), upper, mid+1, high );
 
 		/* If the lower end is the lowest in the alphabet then there is no
 		 * sense testing it. */
 		if ( limitLow ) {
-			out << TABS(level) << "} else {\n";
-			TRANS_GOTO(data[mid].value, level+1) << "\n";
-			out << TABS(level) << "}\n";
+			out << "} else {\n";
+			TRANS_GOTO(data[mid].value) << "\n";
+			out << "}\n";
 		}
 		else {
-			out << TABS(level) << "} else if ( " << GET_KEY() << " >= " << 
+			out << "} else if ( " << GET_KEY() << " >= " << 
 					KEY(data[mid].lowKey) << " ) {\n";
-			TRANS_GOTO(data[mid].value, level+1) << "\n";
-			out << TABS(level) << "}\n";
+			TRANS_GOTO(data[mid].value) << "\n";
+			out << "}\n";
 		}
 	}
 	else {
 		/* Cannot go higher or lower than mid. It's mid or bust. What
 		 * tests to do depends on limits of alphabet. */
 		if ( !limitLow && !limitHigh ) {
-			out << TABS(level) << "if ( " << KEY(data[mid].lowKey) << " <= " << 
+			out << "if ( " << KEY(data[mid].lowKey) << " <= " << 
 					GET_KEY() << " && " << GET_KEY() << " <= " << 
 					KEY(data[mid].highKey) << " ) {\n";
-			TRANS_GOTO(data[mid].value, level+1) << "\n";
-			out << TABS(level) << "}\n";
+			TRANS_GOTO(data[mid].value) << "\n";
+			out << "}\n";
 		}
 		else if ( limitLow && !limitHigh ) {
-			out << TABS(level) << "if ( " << GET_KEY() << " <= " << 
+			out << "if ( " << GET_KEY() << " <= " << 
 					KEY(data[mid].highKey) << " ) {\n";
-			TRANS_GOTO(data[mid].value, level+1) << "\n";
-			out << TABS(level) << "}\n";
+			TRANS_GOTO(data[mid].value) << "\n";
+			out << "}\n";
 		}
 		else if ( !limitLow && limitHigh ) {
-			out << TABS(level) << "if ( " << KEY(data[mid].lowKey) << " <= " << 
+			out << "if ( " << KEY(data[mid].lowKey) << " <= " << 
 					GET_KEY() << " ) {\n";
-			TRANS_GOTO(data[mid].value, level+1) << "\n";
-			out << TABS(level) << "}\n";
+			TRANS_GOTO(data[mid].value) << "\n";
+			out << "}\n";
 		}
 		else {
 			/* Both high and low are at the limit. No tests to do. */
-			out << TABS(level) << "{\n";
-			TRANS_GOTO(data[mid].value, level+1) << "\n";
-			out << TABS(level) << "}\n";
+			out << "{\n";
+			TRANS_GOTO(data[mid].value) << "\n";
+			out << "}\n";
 		}
 	}
 }
@@ -404,7 +251,7 @@ string Goto::CKEY( CondKey key )
 	return ret.str();
 }
 
-void Goto::COND_B_SEARCH( RedTransAp *trans, int level, CondKey lower,
+void Goto::COND_B_SEARCH( RedTransAp *trans, CondKey lower,
 		CondKey upper, int low, int high )
 {
 	/* Get the mid position, staying on the lower end of the range. */
@@ -424,117 +271,144 @@ void Goto::COND_B_SEARCH( RedTransAp *trans, int level, CondKey lower,
 
 	if ( anyLower && anyHigher ) {
 		/* Can go lower and higher than mid. */
-		out << TABS(level) << "if ( " << "ck" << " < " << 
+		out << "if ( " << ck << " < " << 
 				CKEY(midKey) << " ) {\n";
-		COND_B_SEARCH( trans, level+1, lower, midKey-1, low, mid-1 );
-		out << TABS(level) << "} else if ( " << "ck" << " > " << 
+		COND_B_SEARCH( trans, lower, midKey-1, low, mid-1 );
+		out << "} else if ( " << ck << " > " << 
 				CKEY(midKey) << " ) {\n";
-		COND_B_SEARCH( trans, level+1, midKey+1, upper, mid+1, high );
-		out << TABS(level) << "} else {\n";
-		COND_GOTO(midTrans, level+1) << "\n";
-		out << TABS(level) << "}\n";
+		COND_B_SEARCH( trans, midKey+1, upper, mid+1, high );
+		out << "} else {\n";
+		COND_GOTO(midTrans) << "\n";
+		out << "}\n";
 	}
 	else if ( anyLower && !anyHigher ) {
 		/* Can go lower than mid but not higher. */
-		out << TABS(level) << "if ( " << "ck" << " < " << 
+		out << "if ( " << ck << " < " << 
 				CKEY(midKey) << " ) {\n";
-		COND_B_SEARCH( trans, level+1, lower, midKey-1, low, mid-1);
+		COND_B_SEARCH( trans, lower, midKey-1, low, mid-1);
 
 		/* if the higher is the highest in the alphabet then there is no
 		 * sense testing it. */
 		if ( limitHigh ) {
-			out << TABS(level) << "} else {\n";
-			COND_GOTO(midTrans, level+1) << "\n";
-			out << TABS(level) << "}\n";
+			out << "} else {\n";
+			COND_GOTO(midTrans) << "\n";
+			out << "}\n";
 		}
 		else {
-			out << TABS(level) << "} else if ( " << "ck" << " <= " << 
+			out << "} else if ( " << ck << " <= " << 
 					CKEY(midKey) << " ) {\n";
-			COND_GOTO(midTrans, level+1) << "\n";
-			out << TABS(level) << "}\n";
+			COND_GOTO(midTrans) << "\n";
+			out << "}\n";
 		}
 	}
 	else if ( !anyLower && anyHigher ) {
 		/* Can go higher than mid but not lower. */
-		out << TABS(level) << "if ( " << "ck" << " > " << 
+		out << "if ( " << ck << " > " << 
 				CKEY(midKey) << " ) {\n";
-		COND_B_SEARCH( trans, level+1, midKey+1, upper, mid+1, high );
+		COND_B_SEARCH( trans, midKey+1, upper, mid+1, high );
 
 		/* If the lower end is the lowest in the alphabet then there is no
 		 * sense testing it. */
 		if ( limitLow ) {
-			out << TABS(level) << "} else {\n";
-			COND_GOTO(midTrans, level+1) << "\n";
-			out << TABS(level) << "}\n";
+			out << "} else {\n";
+			COND_GOTO(midTrans) << "\n";
+			out << "}\n";
 		}
 		else {
-			out << TABS(level) << "} else if ( " << "ck" << " >= " << 
+			out << "} else if ( " << ck << " >= " << 
 					CKEY(midKey) << " ) {\n";
-			COND_GOTO(midTrans, level+1) << "\n";
-			out << TABS(level) << "}\n";
+			COND_GOTO(midTrans) << "\n";
+			out << "}\n";
 		}
 	}
 	else {
 		/* Cannot go higher or lower than mid. It's mid or bust. What
 		 * tests to do depends on limits of alphabet. */
 		if ( !limitLow && !limitHigh ) {
-			out << TABS(level) << "if ( ck" << " == " << 
+			out << "if ( " << ck << " == " << 
 					CKEY(midKey) << " ) {\n";
-			COND_GOTO(midTrans, level+1) << "\n";
-			out << TABS(level) << "}\n";
+			COND_GOTO(midTrans) << "\n";
+			out << "}\n";
 		}
 		else if ( limitLow && !limitHigh ) {
-			out << TABS(level) << "if ( " << "ck" << " <= " << 
+			out << "if ( " << ck << " <= " << 
 					CKEY(midKey) << " ) {\n";
-			COND_GOTO(midTrans, level+1) << "\n";
-			out << TABS(level) << "}\n";
+			COND_GOTO(midTrans) << "\n";
+			out << "}\n";
 		}
 		else if ( !limitLow && limitHigh ) {
-			out << TABS(level) << "if ( " << CKEY(midKey) << " <= " << 
-					"ck" << " )\n {";
-			COND_GOTO(midTrans, level+1) << "\n";
-			out << TABS(level) << "}\n";
+			out << "if ( " << CKEY(midKey) << " <= " << ck << " )\n {";
+			COND_GOTO(midTrans) << "\n";
+			out << "}\n";
 		}
 		else {
 			/* Both high and low are at the limit. No tests to do. */
-			COND_GOTO(midTrans, level) << "\n";
+			COND_GOTO(midTrans) << "\n";
 		}
 	}
 }
 
 void Goto::STATE_GOTO_ERROR()
 {
-	/* Label the state and bail immediately. */
-	outLabelUsed = true;
-	RedStateAp *state = redFsm->errState;
-	out << "case " << state->id << ":\n";
-	out << "	goto _out;\n";
+	/* Bail out immediately. */
+	out << "	goto " << _again << ";\n";
 }
 
-std::ostream &Goto::STATE_GOTOS()
+void Goto::FROM_STATE_ACTION_EMIT( RedStateAp *state )
 {
+	if ( state->fromStateAction != 0 ) {
+		/* Write every action in the list. */
+		for ( GenActionTable::Iter item = state->fromStateAction->key; item.lte(); item++ ) {
+			ACTION( out, item->value, IlOpts( state->id, false,
+						state->fromStateAction->anyNextStmt() ) );
+			out << "\n";
+		}
+	}
+}
+
+std::ostream &Goto::STATE_CASES()
+{
+	bool eof = redFsm->anyEofActivity() || redFsm->anyNfaStates();
+
 	for ( RedStateList::Iter st = redFsm->stateList; st.lte(); st++ ) {
+		/* Writing code above state gotos. */
+		GOTO_HEADER( st );
+
+		FROM_STATE_ACTION_EMIT( st );
+
+		if ( !noEnd && eof ) {
+			out << 
+				"if ( " << P() << " == " << vEOF() << " ) {\n";
+
+			if ( st->eofTrans != 0 )
+				TRANS_GOTO( st->eofTrans );
+
+			out << 
+				"	goto " << _again << ";\n"
+				"}\n"
+				"else {\n";
+		}
+
 		if ( st == redFsm->errState )
 			STATE_GOTO_ERROR();
 		else {
-			/* Writing code above state gotos. */
-			GOTO_HEADER( st );
-
 			/* Try singles. */
 			if ( st->outSingle.length() > 0 )
 				SINGLE_SWITCH( st );
 
 			/* Default case is to binary search for the ranges, if that fails then */
 			if ( st->outRange.length() > 0 ) {
-				RANGE_B_SEARCH( st, 1, keyOps->minKey, keyOps->maxKey,
+				RANGE_B_SEARCH( st, keyOps->minKey, keyOps->maxKey,
 						0, st->outRange.length() - 1 );
 			}
 
 			/* Write the default transition. */
+			TRANS_GOTO( st->defTrans ) << "\n";
+		}
 
-			out << "{\n";
-			TRANS_GOTO( st->defTrans, 1 ) << "\n";
-			out << "}\n";
+		if ( !noEnd && eof ) {
+			out << 
+				"}\n";
 		}
 	}
 	return out;
@@ -543,11 +417,12 @@ std::ostream &Goto::STATE_GOTOS()
 std::ostream &Goto::TRANSITION( RedCondPair *pair )
 {
 	/* Write the label for the transition so it can be jumped to. */
-	out << "	ctr" << pair->id << ": ";
+	if ( ctrLabel[pair->id].isReferenced )
+		out << "_ctr" << pair->id << ": ";
 
 	/* Destination state. */
 	if ( pair->action != 0 && pair->action->anyCurStateRef() )
-		out << "_ps = " << vCS() << ";";
+		out << ps << " = " << vCS() << ";";
 	out << vCS() << " = " << pair->targ->id << "; ";
 
 	if ( pair->action != 0 ) {
@@ -556,7 +431,7 @@ std::ostream &Goto::TRANSITION( RedCondPair *pair )
 	}
 	else {
 		/* No code to execute, just loop around. */
-		out << "goto _again;\n";
+		out << "goto " << _again << ";\n";
 	}
 	return out;
 }
@@ -745,6 +620,12 @@ void Goto::taNfaPopTrans()
 	nfaPopTrans.finish();
 }
 
+void Goto::EOF_CHECK( ostream &ret )
+{
+	ret << 
+		"	if ( " << P() << " == " << PE() << " )\n"
+		"		goto " << _test_eof << ";\n";
+}
 
 void Goto::GOTO( ostream &ret, int gotoDest, bool inFinish )
 {
@@ -753,7 +634,7 @@ void Goto::GOTO( ostream &ret, int gotoDest, bool inFinish )
 	if ( inFinish && !noEnd )
 		EOF_CHECK( ret );
 
-	ret << "goto _again;";
+	ret << "goto " << _again << ";";
 	
 	ret << CLOSE_GEN_BLOCK();
 }
@@ -767,14 +648,14 @@ void Goto::GOTO_EXPR( ostream &ret, GenInlineItem *ilItem, bool inFinish )
 	if ( inFinish && !noEnd )
 		EOF_CHECK( ret );
 	
-	ret << " goto _again;";
+	ret << " goto " << _again << ";";
 
 	ret << CLOSE_GEN_BLOCK();
 }
 
 void Goto::CURS( ostream &ret, bool inFinish )
 {
-	ret << "(_ps)";
+	ret << "(" << ps << ")";
 }
 
 void Goto::TARGS( ostream &ret, bool inFinish, int targState )
@@ -811,7 +692,7 @@ void Goto::CALL( ostream &ret, int callDest, int targState, bool inFinish )
 	if ( inFinish && !noEnd )
 		EOF_CHECK( ret );
 
-	ret << " goto _again;";
+	ret << " goto " << _again << ";";
 
 	ret << CLOSE_GEN_BLOCK();
 }
@@ -849,7 +730,7 @@ void Goto::CALL_EXPR( ostream &ret, GenInlineItem *ilItem, int targState, bool i
 	if ( inFinish && !noEnd )
 		EOF_CHECK( ret );
 
-	ret << " goto _again;";
+	ret << " goto " << _again << ";";
 
 	ret << CLOSE_GEN_BLOCK();
 }
@@ -883,7 +764,7 @@ void Goto::RET( ostream &ret, bool inFinish )
 	if ( inFinish && !noEnd )
 		EOF_CHECK( ret );
 
-	ret << "goto _again;" << CLOSE_GEN_BLOCK();
+	ret << "goto " << _again << ";" << CLOSE_GEN_BLOCK();
 }
 
 void Goto::NRET( ostream &ret, bool inFinish )
@@ -901,12 +782,197 @@ void Goto::NRET( ostream &ret, bool inFinish )
 
 void Goto::BREAK( ostream &ret, int targState, bool csForced )
 {
-	outLabelUsed = true;
-	ret << OPEN_GEN_BLOCK() << P() << " += 1; " << "goto _out; " << CLOSE_GEN_BLOCK();
+	ret << OPEN_GEN_BLOCK() << P() << " += 1; " << "goto " << _out << "; " << CLOSE_GEN_BLOCK();
 }
 
 void Goto::NBREAK( ostream &ret, int targState, bool csForced )
 {
-	outLabelUsed = true;
-	ret << OPEN_GEN_BLOCK() << P() << " += 1; " << " _nbreak = 1; " << CLOSE_GEN_BLOCK();
+	ret << OPEN_GEN_BLOCK() << P() << " += 1; " << nbreak << " = 1; " << CLOSE_GEN_BLOCK();
+}
+
+void Goto::tableDataPass()
+{
+	if ( type == Loop )
+		taActions();
+
+	taToStateActions();
+	taFromStateActions();
+	taEofActions();
+
+	taNfaTargs();
+	taNfaOffsets();
+	taNfaPushActions();
+	taNfaPopTrans();
+}
+
+void Goto::genAnalysis()
+{
+	/* For directly executable machines there is no required state
+	 * ordering. Choose a depth-first ordering to increase the
+	 * potential for fall-throughs. */
+	redFsm->depthFirstOrdering();
+
+	/* Choose default transitions and the single transition. */
+	redFsm->chooseDefaultSpan();
+		
+	/* Choose single. */
+	redFsm->moveSelectTransToSingle();
+
+	/* If any errors have occured in the input file then don't write anything. */
+	if ( red->id->errorCount > 0 )
+		return;
+
+	/* Anlayze Machine will find the final action reference counts, among other
+	 * things. We will use these in reporting the usage of fsm directives in
+	 * action code. */
+	red->analyzeMachine();
+
+	/* Run the analysis pass over the table data. */
+	setTableState( TableArray::AnalyzePass );
+	tableDataPass();
+
+	/* Switch the tables over to the code gen mode. */
+	setTableState( TableArray::GeneratePass );
+}
+
+void Goto::writeData()
+{
+	if ( type == Loop ) {
+		if ( redFsm->anyActions() )
+			taActions();
+	}
+
+	if ( redFsm->anyToStateActions() )
+		taToStateActions();
+
+	if ( redFsm->anyFromStateActions() )
+		taFromStateActions();
+
+	if ( redFsm->anyEofActions() )
+		taEofActions();
+
+	taNfaTargs();
+	taNfaOffsets();
+	taNfaPushActions();
+	taNfaPopTrans();
+
+	STATE_IDS();
+}
+
+void Goto::writeExec()
+{
+	int maxCtrId = redFsm->nextCondId > redFsm->nextTransId ? redFsm->nextCondId : redFsm->nextTransId;
+	ctrLabel = allocateLabels( ctrLabel, IpLabel::Ctr, maxCtrId );
+
+	out << "{\n";
+
+	DECLARE( INT(), cpc );
+	DECLARE( INT(), ck );
+	DECLARE( INT(), pop_test );
+	DECLARE( INT(), nbreak );
+	DECLARE( INT(), ps, " = 0" );
+	DECLARE( INT(), new_recs );
+	DECLARE( INT(), alt );
+	DECLARE( INDEX( ARR_TYPE( actions ) ), acts );
+	DECLARE( UINT(), nacts );
+
+	out << "\n";
+
+	out << EMIT_LABEL( _resume );
+
+	/* Do we break out on no more input. */
+	bool eof = redFsm->anyEofActivity() || redFsm->anyNfaStates();
+	if ( !noEnd ) {
+		if ( eof ) {
+			out << 
+				"       if ( " << P() << " == " << PE() << " && " << P() << " != " << vEOF() << " )\n"
+				"			goto " << _out << ";\n";
+		}
+		else {
+			out << 
+				"       if ( " << P() << " == " << PE() << " )\n"
+				"			goto " << _out << ";\n";
+		}
+	}
+
+	NFA_PUSH( vCS() ); 
+
+	out <<
+		"	switch ( " << vCS() << " ) {\n";
+		STATE_CASES() <<
+		"	}\n"
+		"\n";
+		TRANSITIONS() <<
+		"\n";
+
+	if ( redFsm->anyRegActions() )
+		EXEC_FUNCS() << "\n";
+
+	out << EMIT_LABEL( _again );
+
+	if ( !noEnd && eof ) {
+		out << 
+			"	if ( " << P() << " == " << vEOF() << " ) {\n"
+			"		if ( " << vCS() << " >= " << FIRST_FINAL_STATE() << " )\n"
+			"			goto " << _out << ";\n"
+			"	}\n"
+			"	else {\n";
+	}
+
+	TO_STATE_ACTIONS();
+
+	if ( redFsm->errState != 0 ) {
+		out << 
+			"	if ( " << vCS() << " != " << redFsm->errState->id << " ) {\n";
+	}
+
+	out << 
+		"	" << P() << " += 1;\n"
+		"	goto " << _resume << ";\n";
+
+	if ( redFsm->errState != 0 ) {
+		out << 
+			"	}\n";
+	}
+
+	if ( !noEnd && eof ) {
+		out <<
+			"	}\n";
+	}
+
+	if ( redFsm->anyNfaStates() ) {
+		out <<
+			"	if ( nfa_len == 0 )\n"
+			"		goto " << _out << ";\n"
+			"\n"
+			"	nfa_count += 1;\n"
+			"	nfa_len -= 1;\n"
+			"	" << P() << " = nfa_bp[nfa_len].p;\n"
+			;
+
+		if ( redFsm->bAnyNfaPops ) {
+			NFA_FROM_STATE_ACTION_EXEC();
+
+			NFA_POP_TEST_EXEC();
+
+			out <<
+				"	if ( " << pop_test << " )\n"
+				"		" << vCS() << " = nfa_bp[nfa_len].state;\n"
+				"	else\n"
+				"		" << vCS() << " = " << ERROR_STATE() << ";\n";
+		}
+		else {
+			out <<
+				"	" << vCS() << " = nfa_bp[nfa_len].state;\n";
+
+		}
+
+		NFA_POST_POP();
+
+		out << "goto " << _resume << ";\n";
+	}
+
+	out << EMIT_LABEL( _out );
+
+	out << "}\n";
 }
