@@ -1,5 +1,5 @@
 /*
- * Copyright 2001-2014 Adrian Thurston <thurston@colm.net>
+ * Copyright 2001-2018 Adrian Thurston <thurston@colm.net>
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to
@@ -28,49 +28,7 @@
 #include "parsedata.h"
 #include "inputdata.h"
 
-void SwitchExpGoto::tableDataPass()
-{
-	taToStateActions();
-	taFromStateActions();
-	taEofActions();
-
-	taNfaTargs();
-	taNfaOffsets();
-	taNfaPushActions();
-	taNfaPopTrans();
-}
-
-void SwitchExpGoto::genAnalysis()
-{
-	/* For directly executable machines there is no required state
-	 * ordering. Choose a depth-first ordering to increase the
-	 * potential for fall-throughs. */
-	redFsm->depthFirstOrdering();
-
-	/* Choose default transitions and the single transition. */
-	redFsm->chooseDefaultSpan();
-		
-	/* Choose single. */
-	redFsm->moveSelectTransToSingle();
-
-	/* If any errors have occured in the input file then don't write anything. */
-	if ( red->id->errorCount > 0 )
-		return;
-	
-	/* Anlayze Machine will find the final action reference counts, among other
-	 * things. We will use these in reporting the usage of fsm directives in
-	 * action code. */
-	red->analyzeMachine();
-
-	/* Run the analysis pass over the table data. */
-	setTableState( TableArray::AnalyzePass );
-	tableDataPass();
-
-	/* Switch the tables over to the code gen mode. */
-	setTableState( TableArray::GeneratePass );
-}
-
-std::ostream &SwitchExpGoto::EXEC_ACTIONS()
+std::ostream &GotoExp::EXEC_FUNCS()
 {
 	/* Loop the actions. */
 	for ( GenActionTableMap::Iter redAct = redFsm->actionMap; redAct.lte(); redAct++ ) {
@@ -79,7 +37,7 @@ std::ostream &SwitchExpGoto::EXEC_ACTIONS()
 			out << "f" << redAct->actListId << ":\n";
 
 			if ( redFsm->anyRegNbreak() )
-				out << "_nbreak = 0;\n";
+				out << nbreak << " = 0;\n";
 
 			/* Write each action in the list of action items. */
 			for ( GenActionTable::Iter item = redAct->key; item.lte(); item++ )
@@ -87,13 +45,12 @@ std::ostream &SwitchExpGoto::EXEC_ACTIONS()
 
 			if ( redFsm->anyRegNbreak() ) {
 				out << 
-					"	if ( _nbreak == 1 )\n"
-					"		goto _out;\n";
-				outLabelUsed = true;
+					"	if ( " << nbreak << " == 1 )\n"
+					"		goto " << _out << ";\n";
 			}
 
 
-			out << "\n\tgoto _again;\n";
+			out << "goto " << _again << ";\n";
 		}
 	}
 	return out;
@@ -101,7 +58,7 @@ std::ostream &SwitchExpGoto::EXEC_ACTIONS()
 
 /* Write out the function switch. This switch is keyed on the values
  * of the func index. */
-std::ostream &SwitchExpGoto::TO_STATE_ACTION_SWITCH()
+std::ostream &GotoExp::TO_STATE_ACTION_SWITCH()
 {
 	/* Loop the actions. */
 	for ( GenActionTableMap::Iter redAct = redFsm->actionMap; redAct.lte(); redAct++ ) {
@@ -113,7 +70,7 @@ std::ostream &SwitchExpGoto::TO_STATE_ACTION_SWITCH()
 			for ( GenActionTable::Iter item = redAct->key; item.lte(); item++ )
 				ACTION( out, item->value, IlOpts( 0, false, false ) );
 
-			out << "\n\t" << CEND() << "}\n";
+			out << "\n\t" << CEND() << "\n}\n";
 		}
 	}
 
@@ -122,7 +79,7 @@ std::ostream &SwitchExpGoto::TO_STATE_ACTION_SWITCH()
 
 /* Write out the function switch. This switch is keyed on the values
  * of the func index. */
-std::ostream &SwitchExpGoto::FROM_STATE_ACTION_SWITCH()
+std::ostream &GotoExp::FROM_STATE_ACTION_SWITCH()
 {
 	/* Loop the actions. */
 	for ( GenActionTableMap::Iter redAct = redFsm->actionMap; redAct.lte(); redAct++ ) {
@@ -134,14 +91,14 @@ std::ostream &SwitchExpGoto::FROM_STATE_ACTION_SWITCH()
 			for ( GenActionTable::Iter item = redAct->key; item.lte(); item++ )
 				ACTION( out, item->value, IlOpts( 0, false, false ) );
 
-			out << "\n\t" << CEND() << "}\n";
+			out << "\n\t" << CEND() << "\n}\n";
 		}
 	}
 
 	return out;
 }
 
-std::ostream &SwitchExpGoto::EOF_ACTION_SWITCH()
+std::ostream &GotoExp::EOF_ACTION_SWITCH()
 {
 	/* Loop the actions. */
 	for ( GenActionTableMap::Iter redAct = redFsm->actionMap; redAct.lte(); redAct++ ) {
@@ -153,14 +110,14 @@ std::ostream &SwitchExpGoto::EOF_ACTION_SWITCH()
 			for ( GenActionTable::Iter item = redAct->key; item.lte(); item++ )
 				ACTION( out, item->value, IlOpts( 0, true, false ) );
 
-			out << "\n\t" << CEND() << "}\n";
+			out << "\n\t" << CEND() << "\n}\n";
 		}
 	}
 
 	return out;
 }
 
-unsigned int SwitchExpGoto::TO_STATE_ACTION( RedStateAp *state )
+unsigned int GotoExp::TO_STATE_ACTION( RedStateAp *state )
 {
 	int act = 0;
 	if ( state->toStateAction != 0 )
@@ -168,7 +125,7 @@ unsigned int SwitchExpGoto::TO_STATE_ACTION( RedStateAp *state )
 	return act;
 }
 
-unsigned int SwitchExpGoto::FROM_STATE_ACTION( RedStateAp *state )
+unsigned int GotoExp::FROM_STATE_ACTION( RedStateAp *state )
 {
 	int act = 0;
 	if ( state->fromStateAction != 0 )
@@ -176,7 +133,7 @@ unsigned int SwitchExpGoto::FROM_STATE_ACTION( RedStateAp *state )
 	return act;
 }
 
-unsigned int SwitchExpGoto::EOF_ACTION( RedStateAp *state )
+unsigned int GotoExp::EOF_ACTION( RedStateAp *state )
 {
 	int act = 0;
 	if ( state->eofAction != 0 )
@@ -184,7 +141,7 @@ unsigned int SwitchExpGoto::EOF_ACTION( RedStateAp *state )
 	return act;
 }
 
-void SwitchExpGoto::NFA_PUSH_ACTION( RedNfaTarg *targ )
+void GotoExp::NFA_PUSH_ACTION( RedNfaTarg *targ )
 {
 	int act = 0;
 	if ( targ->push != 0 )
@@ -192,7 +149,7 @@ void SwitchExpGoto::NFA_PUSH_ACTION( RedNfaTarg *targ )
 	nfaPushActions.value( act );
 }
 
-void SwitchExpGoto::NFA_POP_TEST( RedNfaTarg *targ )
+void GotoExp::NFA_POP_TEST( RedNfaTarg *targ )
 {
 	int act = 0;
 	if ( targ->popTest != 0 )
@@ -201,26 +158,7 @@ void SwitchExpGoto::NFA_POP_TEST( RedNfaTarg *targ )
 }
 
 
-void SwitchExpGoto::writeData()
-{
-	if ( redFsm->anyToStateActions() )
-		taToStateActions();
-
-	if ( redFsm->anyFromStateActions() )
-		taFromStateActions();
-
-	if ( redFsm->anyEofActions() )
-		taEofActions();
-
-	taNfaTargs();
-	taNfaOffsets();
-	taNfaPushActions();
-	taNfaPopTrans();
-
-	STATE_IDS();
-}
-
-void SwitchExpGoto::NFA_FROM_STATE_ACTION_EXEC()
+void GotoExp::NFA_FROM_STATE_ACTION_EXEC()
 {
 	if ( redFsm->anyFromStateActions() ) {
 		out <<
@@ -231,37 +169,8 @@ void SwitchExpGoto::NFA_FROM_STATE_ACTION_EXEC()
 	}
 }
 
-void SwitchExpGoto::writeExec()
+void GotoExp::FROM_STATE_ACTIONS()
 {
-	testEofUsed = false;
-	outLabelUsed = false;
-
-	out << "	{\n";
-
-	if ( redFsm->anyRegCurStateRef() )
-		out << "	int _ps = 0;\n";
-
-	if ( redFsm->anyRegNbreak() ) {
-		out << 
-			"	int _nbreak;\n";
-	}
-
-	if ( !noEnd ) {
-		testEofUsed = true;
-		out << 
-			"	if ( " << P() << " == " << PE() << " )\n"
-			"		goto _test_eof;\n";
-	}
-
-	if ( redFsm->errState != 0 ) {
-		outLabelUsed = true;
-		out << 
-			"	if ( " << vCS() << " == " << redFsm->errState->id << " )\n"
-			"		goto _out;\n";
-	}
-
-	out << "_resume:\n";
-
 	if ( redFsm->anyFromStateActions() ) {
 		out <<
 			"	switch ( " << ARR_REF( fromStateActions ) << "[" << vCS() << "] ) {\n";
@@ -269,22 +178,10 @@ void SwitchExpGoto::writeExec()
 			"	}\n"
 			"\n";
 	}
+}
 
-	NFA_PUSH();
-
-	out << 
-		"	switch ( " << vCS() << " ) {\n";
-		STATE_GOTOS() <<
-		"	}\n"
-		"\n";
-		TRANSITIONS() << 
-		"\n";
-
-	if ( redFsm->anyRegActions() )
-		EXEC_ACTIONS() << "\n";
-
-	out << "_again:\n";
-
+void GotoExp::TO_STATE_ACTIONS()
+{
 	if ( redFsm->anyToStateActions() ) {
 		out <<
 			"	switch ( " << ARR_REF( toStateActions ) << "[" << vCS() << "] ) {\n";
@@ -292,65 +189,20 @@ void SwitchExpGoto::writeExec()
 			"	}\n"
 			"\n";
 	}
+}
 
-	if ( redFsm->errState != 0 ) {
-		outLabelUsed = true;
-		out << 
-			"	if ( " << vCS() << " == " << redFsm->errState->id << " )\n"
-			"		goto _out;\n";
-	}
-
-	if ( !noEnd ) {
-		out << 
-			"	" << P() << "+= 1;\n"
-			"	if ( " << P() << " != " << PE() << " )\n"
-			"		goto _resume;\n";
-	}
-	else {
-		out << 
-			"	" << P() << " += 1;\n"
-			"	goto _resume;\n";
-	}
-
-	if ( testEofUsed )
-		out << "	_test_eof: {}\n";
-
-	if ( redFsm->anyEofTrans() || redFsm->anyEofActions() ) {
-		out <<
-			"	if ( " << P() << " == " << vEOF() << " )\n"
-			"	{\n";
-
-		if ( redFsm->anyEofTrans() ) {
-			out <<
-				"	switch ( " << vCS() << " ) {\n";
-
-			for ( RedStateList::Iter st = redFsm->stateList; st.lte(); st++ ) {
-				if ( st->eofTrans != 0 ) {
-					RedCondPair *cond = st->eofTrans->outCond( 0 );
-					out << "	case " << st->id << ": goto ctr" << cond->id << ";\n";
-				}
-			}
+void GotoExp::REG_ACTIONS()
+{
 
-			out <<
-				"	}\n";
-		}
+}
 
-		if ( redFsm->anyEofActions() ) {
-			out <<
-				"	switch ( " << ARR_REF( eofActions ) << "[" << vCS() << "] ) {\n";
-				EOF_ACTION_SWITCH() <<
-				"	}\n";
-		}
-
+void GotoExp::EOF_ACTIONS()
+{
+	if ( redFsm->anyEofActions() ) {
 		out <<
-			"	}\n"
-			"\n";
+			"	switch ( " << ARR_REF( eofActions ) << "[" << vCS() << "] ) {\n";
+			EOF_ACTION_SWITCH() <<
+			"	}\n";
 	}
-
-	if ( outLabelUsed )
-		out << "	_out: {}\n";
-
-	NFA_POP();
 
-	out << "	}\n";
 }
