@@ -1,5 +1,5 @@
 /*
- * Copyright 2008 Adrian Thurston <thurston@colm.net>
+ * Copyright 2008-2018 Adrian Thurston <thurston@colm.net>
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to
@@ -34,6 +34,7 @@ struct CondSpace;
 struct CondAp;
 struct ActionTable;
 struct Section;
+struct LangFuncs;
 
 void translatedHostData( ostream &out, const string &data );
 
@@ -169,18 +170,20 @@ struct InputData
 :
 	public FsmGbl
 {
-	InputData()
+	InputData( const HostLang *hostLang,
+			struct colm_sections *frontendSections, struct colm_sections *rlhcSections )
 	: 
+		FsmGbl(hostLang),
+		frontendSections(frontendSections),
+		rlhcSections(rlhcSections),
 		inputFileName(0),
 		outputFileName(0),
-		commFileName(0),
 		nextMachineId(0),
 		inStream(0),
 		outStream(0),
 		outFilter(0),
 		curItem(0),
 		lastFlush(0),
-		hostLang(&hostLangC),
 		codeStyle(GenBinaryLoop),
 		dotGenPd(0),
 		machineSpec(0),
@@ -189,10 +192,12 @@ struct InputData
 		noLineDirectives(false),
 		maxTransitions(LONG_MAX),
 		numSplitPartitions(0),
-		rubyImpl(MRI),
+		rlhc(false),
 		rlhcShowCmd(false),
 		noIntermediate(false),
 		frontendSpecified(false),
+		backendSpecified(false),
+		featureSpecified(false),
 		saveTemps(false),
 		condsCheckDepth(-1),
 		transSpanDepth(6),
@@ -202,25 +207,23 @@ struct InputData
 		histogramFn(0),
 		histogram(0),
 		input(0),
-		forceLibRagel(false)
+		forceVar(false)
 	{}
 
 	~InputData();
 
 	void usage();
 	void version();
-	void showHostLangNames();
-	void showHostLangArgs();
 	void showFrontends();
 	void showBackends();
-	void showStyles();
 
+	struct colm_sections *frontendSections;
+	struct colm_sections *rlhcSections;
 	std::string dirName;
 
 	/* The name of the root section, this does not change during an include. */
 	const char *inputFileName;
 	const char *outputFileName;
-	const char *commFileName;
 
 	string comm;
 
@@ -240,8 +243,6 @@ struct InputData
 	InputItem *curItem;
 	InputItem *lastFlush;
 
-	const HostLang *hostLang;
-
 	/* Ragel-6 frontend. */
 	ParserDict parserDict;
 	ParserList parserList;
@@ -269,15 +270,17 @@ struct InputData
 	long maxTransitions;
 	int numSplitPartitions;
 
-	/* Target ruby impl */
-	RubyImplEnum rubyImpl;
-
+	bool rlhc;
 	bool rlhcShowCmd;
 	bool noIntermediate;
 
 	bool frontendSpecified;
 	RagelFrontend frontend;
 
+	bool backendSpecified;
+
+	bool featureSpecified;
+
 	bool saveTemps;
 	long condsCheckDepth;
 	long transSpanDepth;
@@ -293,6 +296,8 @@ struct InputData
 
 	Vector<const char**> streamFileNames;
 
+	bool forceVar;
+
 	void verifyWriteHasData( InputItem *ii );
 	void verifyWritesHaveData();
 
@@ -308,9 +313,6 @@ struct InputData
 	void prepareSingleMachine();
 	void prepareAllMachines();
 
-	void cDefaultFileName( const char *inputFile );
-	void asmDefaultFileName( const char *inputFile );
-
 	void writeOutput( InputItem *ii );
 	void writeLanguage( std::ostream &out );
 
@@ -318,6 +320,7 @@ struct InputData
 
 	void parseKelbt();
 	void processDot();
+	void processCodeEarly();
 
 	void writeDot( std::ostream &out );
 
@@ -329,7 +332,6 @@ struct InputData
 	void terminateParser( Parser6 *parser );
 	void terminateAllParsers();
 
-	void runRlhc();
 	void processKelbt();
 	void processColm();
 	bool processReduce();
@@ -338,10 +340,14 @@ struct InputData
 
 	char *readInput( const char *inputFileName );
 
-	bool forceLibRagel;
-
 	const char **makeIncludePathChecks( const char *curFileName, const char *fileName );
 	std::ifstream *tryOpenInclude( const char **pathChecks, long &found );
+	int main( int argc, const char **argv );
+
+	void wait( const char *what, pid_t pid );
+
+	int rlhcRun( int argc, const char **argv );
+	int rlhcMain( int argc, const char **argv );
 };
 
 
