@@ -1,5 +1,5 @@
 /*
- * Copyright 2001-2014 Adrian Thurston <thurston@colm.net>
+ * Copyright 2001-2018 Adrian Thurston <thurston@colm.net>
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to
@@ -37,16 +37,34 @@ class IpGoto
 {
 public:
 	IpGoto( const CodeGenArgs &args ) 
-			: Goto(args) {}
+	:
+		Goto( args, Ip ),
+		stLabel(0),
+		eofLabel(0),
+		ctrLabel(0)
+	{}
 
 	std::ostream &EXIT_STATES();
-	std::ostream &TRANS_GOTO( RedTransAp *trans, int level );
-	std::ostream &COND_GOTO( RedCondPair *trans, int level );
+	std::ostream &TRANS_GOTO( RedTransAp *trans );
+	std::ostream &COND_GOTO( RedCondPair *trans );
 	std::ostream &FINISH_CASES();
 	std::ostream &AGAIN_CASES();
 	std::ostream &STATE_GOTOS();
 	std::ostream &STATE_GOTO_CASES();
 
+	/* unused. */
+	virtual std::ostream &ACTION_SWITCH() { return out; }
+	virtual std::ostream &EXEC_FUNCS() { return out; }
+	virtual std::ostream &TO_STATE_ACTION_SWITCH() { return out; }
+	virtual std::ostream &FROM_STATE_ACTION_SWITCH() { return out; }
+	virtual std::ostream &EOF_ACTION_SWITCH() { return out; }
+
+	/* Unused */
+	virtual void FROM_STATE_ACTIONS() {}
+	virtual void TO_STATE_ACTIONS() {}
+	virtual void REG_ACTIONS() {}
+	virtual void EOF_ACTIONS() {}
+
 	void GOTO( ostream &ret, int gotoDest, bool inFinish );
 	void CALL( ostream &ret, int callDest, int targState, bool inFinish );
 	void NCALL( ostream &ret, int callDest, int targState, bool inFinish );
@@ -89,6 +107,10 @@ protected:
 	void NFA_PUSH( RedStateAp *state );
 
 	void tableDataPass();
+
+	IpLabel *stLabel;
+	IpLabel *eofLabel;
+	IpLabel *ctrLabel;
 };
 
 namespace C
