@@ -1,5 +1,5 @@
 /*
- * Copyright 2015 Adrian Thurston <thurston@colm.net>
+ * Copyright 2015-2018 Adrian Thurston <thurston@colm.net>
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to
@@ -29,7 +29,6 @@
 
 using std::endl;
 using std::ifstream;
-extern colm_sections rlparse_object;
 
 void TopLevel::loadMachineName( string data )
 {
@@ -106,7 +105,7 @@ void TopLevel::include( const InputLoc &incLoc, bool fileSpecified, string fileN
 
 	const char *inclSectionName = machine.c_str();
 
-	/* Implement defaults for the input file and section name. */
+	/* Default the section name to the current section name. */
 	if ( inclSectionName == 0 )
 		inclSectionName = sectionName.c_str();
 
@@ -155,7 +154,7 @@ void TopLevel::include( const InputLoc &incLoc, bool fileSpecified, string fileN
 	targetMachine = sectionName.c_str();
 	searchMachine = machine.c_str();
 
-	reduceFile( includeChecks[found], false );
+	// reduceFile( includeChecks[found] );
 
 //	if ( includePass.incItems.length() == 0 ) {
 //		pd->id->error(incLoc) << "could not find machine " << machine <<
@@ -175,13 +174,6 @@ void TopLevel::include( const InputLoc &incLoc, bool fileSpecified, string fileN
 	searchMachine = searchMachine0;
 }
 
-void TopLevel::importFile( std::string file )
-{
-	isImport = true;
-	reduceFile( file.c_str(), true );
-	isImport = false;
-}
-
 void TopLevel::import( const InputLoc &loc, std::string name, Literal *literal )
 {
 	MachineDef *machineDef = new MachineDef(
@@ -203,28 +195,24 @@ void TopLevel::import( const InputLoc &loc, std::string name, Literal *literal )
 	machineDef->join->loc = loc;
 }
 
-void TopLevel::reduceFile( const char *inputFileName, bool import )
+void TopLevel::reduceFile( const char *cmd, const char *inputFileName )
 {
-	char idstr[64], imstr[64];
-	sprintf( idstr, "%d", includeDepth );
-	sprintf( imstr, "%d", import );
-
-	const char *argv[7];
-	argv[0] = "rlparse";
+	const int baseN = 2;
+	const char **argv = new const char*[baseN + id->includePaths.length() + 1];
+	argv[0] = cmd;
 	argv[1] = inputFileName;
-	argv[2] = imstr; // Import 
-	argv[3] = idstr; // IncludeDepth
-	argv[4] = targetMachine == 0 ? "" : targetMachine;
-	argv[5] = searchMachine == 0 ? "" : searchMachine;
-	argv[6] = 0;
+	for ( int i = 0; i < id->includePaths.length(); i++ )
+		argv[baseN + i] = id->includePaths.data[i];
+	argv[baseN + id->includePaths.length()] = 0;
 
 	const char *prevCurFileName = curFileName;
 	curFileName = inputFileName;
 
-	colm_program *program = colm_new_program( &rlparse_object );
+	colm_program *program = colm_new_program( frontendSections );
 	colm_set_debug( program, 0 );
+	colm_set_reduce_clean( program, 0 );
 	colm_set_reduce_ctx( program, this );
-	colm_run_program( program, 6, argv );
+	colm_run_program( program, baseN + id->includePaths.length(), argv );
 	id->streamFileNames.append( colm_extract_fns( program ) );
 
 	int length = 0;
@@ -237,6 +225,6 @@ void TopLevel::reduceFile( const char *inputFileName, bool import )
 	colm_delete_program( program );
 
 	curFileName = prevCurFileName;
+
+	delete[] argv;
 }
-
-
